/*
 * Copyright (c) 2016-2017,2019-2022 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Tue Aug 23 2016
 */
#include "aclibmrsim.h"

#include <math.h>
#include <stdlib.h>

#include <Eigen/Geometry>

#include "private.h"


/* --- mrsim_gtmrp_body ---------------------------------------------------- */

/*
 * cx, cy, cz: Position of center of rotors wrt center of mass (m)
 * armlen: Distance of rotors to the center (m)
 * mass: Total mass (kg)
 * mbodyw, mbodyh: Main mass body width and height (m)
 * mmotor: Motor mass (kg)
 * rx, ry: X, Y rotation of the first rotor thrust vector (rad)
 * rz: Z spin direction of the first rotor (1: ccw, -1: cw)
 * cf: Propeller thrust coefficient (N.s²)
 * ct: Propeller torque coefficient (N.m.s²)
 */
int
mrsim_gtmrp_body(const struct mrsim_s *mrsim,
                 unsigned short rotors,
                 double cx, double cy, double cz,
                 double armlen,
                 double mass, double mbodyw, double mbodyh, double mmotor,
                 double rx, double ry, double rz,
                 double cf, double ct)
{
  struct mrsim_param_s *param = mrsim_params(mrsim);
  using namespace Eigen;

  Map< Matrix<double, 6, Dynamic, RowMajor> >
    G_(param->body.G, 6, param->body.rotors);
  Map< Matrix<double, 3, 3, RowMajor> > J_(param->body.J);
  double bmass, izz;
  Vector3d z, p;
  int i, sign;

  /* allocation matrix */
  if (rotors > param->body.rotors) { errno = EINVAL; return -1; }
  param->motor.Cf = cf;
  param->motor.Ct = ct;
  param->motor.Ct = ct;

  for(i = 0, sign = 1; i < rotors; i++, sign = -sign) {
    /* thrust direction */
    z =
      (AngleAxisd(2 * i * M_PI / rotors, Vector3d::UnitZ())
       * AngleAxisd(sign * rx, Vector3d::UnitX())
       * AngleAxisd(ry, Vector3d::UnitY())).matrix().col(2);

    /* thrust position */
    p =
      armlen *
      AngleAxisd(2 * i * M_PI / rotors, Vector3d::UnitZ()).matrix().col(0)
      + Vector3d(cx, cy, cz);

    /* wrench */
    G_.col(i) <<
      cf * z,
      cf * p.cross(z) - sign * rz * ct * z;

    /* enable motor */
    mrsim_motor(mrsim, i)->disabled = 0;
  }
  for(; i < param->body.rotors; i++) {
    G_.col(i) << Vector3d::Zero(), Vector3d::Zero();
    mrsim_motor(mrsim, i)->disabled = 1;
  }


  /* inertia matrix */
  param->body.m = mass;

  bmass = mass - rotors * mmotor;
  if (bmass <= 0.) { errno = EINVAL; return -1; }

  /* main body (rectangular cuboid) */
  J_ = Vector3d(
    1/12. * bmass * (mbodyw * mbodyw + mbodyh * mbodyh),
    1/12. * bmass * (mbodyw * mbodyw + mbodyh * mbodyh),
    1/6. * bmass * mbodyw * mbodyw).asDiagonal();

  /* motors (circular loop) */
  izz = rotors * mmotor * armlen * armlen;
  J_ += Vector3d(
    izz / 2.,
    izz / 2.,
    izz).asDiagonal();

  return 0;
}


/* --- mrsim_motor_wrench -------------------------------------------------- */

struct mrsimstate_wrench_s
mrsim_motor_wrench(const struct mrsim_s *mrsim, uint8_t index,
                   double zaxis[3])
{
  using namespace Eigen;

  const struct mrsim_param_s *params = mrsim_params(mrsim);
  const struct mrsimstate_motor_s *motor = mrsim_motor(mrsim, index);

  struct mrsimstate_wrench_s wrench;
  Map<Vector3d> F(wrench.thrust);
  Map<Vector3d> T(wrench.torque);
  Map<Vector3d> z(zaxis);

  double w2 = copysign(motor->w * motor->w, motor->spin);

  /* aerodynamic thrust & torque */
  F = z * params->motor.Cf * w2;
  T = - z * params->motor.Ct * w2;

  /* vibrations - heuristic model matching observed real behaviour */
  F +=
    Quaterniond::FromTwoVectors(Vector3d::UnitZ(), z) * Vector3d(
      /* imbalance radial force */
      std::cos(motor->a) * params->motor.Vfxy,
      std::sin(motor->a) * params->motor.Vfxy,
      /* axial vibration */
      std::cos(2 * motor->a) * params->motor.Vfz) * params->motor.Cf * w2;

  T +=
    Quaterniond::FromTwoVectors(Vector3d::UnitZ(), z) * Vector3d(
      /* axial vibration */
      std::sin(2 * motor->a) * params->motor.Vtxy,
      std::cos(2 * motor->a) * params->motor.Vtxy,
      /* imbalance radial force */
      std::cos(motor->a) * params->motor.Vtz) * params->motor.Ct * w2;

  return wrench;
}


/* --- mrsim_update_wrench ------------------------------------------------- */

void
mrsim_update_wrench(const struct mrsim_s *mrsim)
{
  using namespace Eigen;

  const struct mrsim_param_s *params = mrsim_params(mrsim);

  unsigned int j;
  Map<Vector3d> F(mrsim->wrench->thrust);
  Map<Vector3d> T(mrsim->wrench->torque);
  Vector3d Cfz, Ctz;
  double w2;

  F << 0., 0., 0.;
  T << 0., 0., 0.;

  for(j = 0; j < mrsim->params->body.rotors; j++) {
    const struct mrsimstate_motor_s *motor = mrsim_motor(mrsim, j);

    /* signed squared velocity #j */
    w2 = copysign(motor->w * motor->w, motor->spin);

    /* Z axis #j */
    Cfz <<
      params->body.G[0*params->body.rotors + j],
      params->body.G[1*params->body.rotors + j],
      params->body.G[2*params->body.rotors + j];

    Ctz = Vector3d(
      params->body.G[3*params->body.rotors + j],
      params->body.G[4*params->body.rotors + j],
      params->body.G[5*params->body.rotors + j]);

    /* aerodynamic thrust & torque */
    F += Cfz * w2;
    T += Ctz * w2;

    /* vibrations - heuristic model matching observed real behaviour */
    F +=
      Quaterniond::FromTwoVectors(Vector3d::UnitZ(), Cfz) * Vector3d(
        /* imbalance radial force */
        std::cos(motor->a) * params->motor.Vfxy,
        std::sin(motor->a) * params->motor.Vfxy,
        /* axial vibration */
        std::cos(2 * motor->a) * params->motor.Vfz) * Cfz.norm() * w2;
    T +=
      Quaterniond::FromTwoVectors(Vector3d::UnitZ(), Cfz) * Vector3d(
        /* axial vibration */
        std::sin(2 * motor->a) * params->motor.Vtxy,
        std::cos(2 * motor->a) * params->motor.Vtxy,
        /* imbalance radial force */
        std::cos(motor->a) * params->motor.Vtz) * Ctz.norm() * w2;
  }
}


/* --- mrsim_update_imu ---------------------------------------------------- */

void
mrsim_update_imu(const struct mrsim_s *mrsim)
{
  using namespace Eigen;

  const Vector3d g(0., 0., - mrsim->params->body.g);
  const Vector3d B(mrsim->params->body.B[0],
                   mrsim->params->body.B[1],
                   mrsim->params->body.B[2]);
  Quaternion<double> q(
    mrsim->body->qw, mrsim->body->qx, mrsim->body->qy, mrsim->body->qz);

  Vector3d w(mrsim->body->wx, mrsim->body->wy, mrsim->body->wz);
  Vector3d a(mrsim->body->ax, mrsim->body->ay, mrsim->body->az);

  Map<Vector3d> wb(mrsim->imu->w);
  Map<Vector3d> ab(mrsim->imu->a);
  Map<Vector3d> Bb(mrsim->imu->B);

  wb = q.conjugate() * w + Vector3d(
    mrsim_gnoise(mrsim->params->state.wstddev),
    mrsim_gnoise(mrsim->params->state.wstddev),
    mrsim_gnoise(mrsim->params->state.wstddev));

  ab = q.conjugate() * (a - g) + Vector3d(
    mrsim_gnoise(mrsim->params->state.astddev),
    mrsim_gnoise(mrsim->params->state.astddev),
    mrsim_gnoise(mrsim->params->state.astddev));

  Bb = q.conjugate() * B + Vector3d(
    mrsim_gnoise(mrsim->params->state.Bstddev),
    mrsim_gnoise(mrsim->params->state.Bstddev),
    mrsim_gnoise(mrsim->params->state.Bstddev));
}


/* --- mrsim_update_body --------------------------------------------------- */

void
mrsim_update_body(const struct mrsim_s *mrsim)
{
  const Eigen::Matrix3d J(mrsim->params->body.J);
  const Eigen::Vector3d thrust(mrsim->wrench->thrust);
  const Eigen::Vector3d torque(mrsim->wrench->torque);
  const Eigen::Vector3d g(0, 0, - mrsim->params->body.g);
  const double dt = mrsim->ts->dt;

  /* current state */
  Eigen::Vector3d p(mrsim->body->px, mrsim->body->py, mrsim->body->pz);
  Eigen::Quaternion<double> q(
    mrsim->body->qw, mrsim->body->qx, mrsim->body->qy, mrsim->body->qz);
  Eigen::Vector3d v(mrsim->body->vx, mrsim->body->vy, mrsim->body->vz);
  Eigen::Vector3d w(mrsim->body->wx, mrsim->body->wy, mrsim->body->wz);
  Eigen::Vector3d a = g + q * thrust / mrsim->params->body.m;


  /* update position */
  p += v * dt + a * dt*dt/2;

  /* update linear velocity */
  v += a * dt;

  /* update angular velocity */
  Eigen::Vector3d wb, dwb;
  wb = q.conjugate() * w;
  dwb = J.inverse()*((J*wb).cross(wb) + torque);
  wb += dwb * dt;
  w = q * wb;

  /* update attitude */
  Eigen::Quaternion<double> dq;
  double a2 = dt * dt * w.squaredNorm();
  if (a2 < 1e-1) {
    dq.w() = 1 - a2/8 /*std::cos(a/2)*/;
    dq.vec() = (0.5 - a2/48 /*std::sin(a/2)/a*/) * dt * w;
  } else {
    double a = std::sqrt(a2);
    dq.w() = std::cos(a/2);
    dq.vec() = std::sin(a/2)/a * dt * w;
  }
  q = dq * q;


  /* update state */
  mrsim->body->px = p(0);
  mrsim->body->py = p(1);
  mrsim->body->pz = p(2);
  mrsim->body->qw = q.w();
  mrsim->body->qx = q.vec()(0);
  mrsim->body->qy = q.vec()(1);
  mrsim->body->qz = q.vec()(2);

  mrsim->body->vx = v(0);
  mrsim->body->vy = v(1);
  mrsim->body->vz = v(2);
  mrsim->body->wx = w(0);
  mrsim->body->wy = w(1);
  mrsim->body->wz = w(2);

  mrsim->body->ax = a(0);
  mrsim->body->ay = a(1);
  mrsim->body->az = a(2);
}


/* --- mrsim_gnoise -------------------------------------------------------- */

double
mrsim_gnoise(double stddev)
{
  double u1, u2;
  double s;

  do {
    u1 = 2 * drand48() - 1; /* [-1; 1] */
    u2 = 2 * drand48() - 1;

    s = u1*u1 + u2*u2;
  } while(s >= 1.);

  return sqrt(-2. * log(s)/s) * stddev * u1;
}
