/*
 * Copyright (c) 2016,2019,2022 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Tue Sep 20 2016
 */
#include "aclibmrsim.h"

#define _XOPEN_SOURCE 600 /* linux does not have ptsname(3) otherwise */
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/uio.h>

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include "private.h"


/* --- mrsim_open_pty ------------------------------------------------------ */

/* Open a pseudo terminal, create a symlink */
int
mrsim_open_pty(const struct mrsim_s *mrsim, const char *device)
{
  int fd, s;

  /* close previous pty if any */
  mrsim_close_pty(mrsim->pty);

  /* open non-blocking */
  fd = open("/dev/ptmx", O_RDWR | O_NOCTTY);
  if (fd < 0) return errno;

  s = fcntl(fd, F_GETFL);
  if (s == -1) goto err;
  if (fcntl(fd, F_SETFL, s | O_NONBLOCK) == -1) goto err;

  /* create symlink */
  unlink(device);
  if (symlink(ptsname(fd), device)) goto err;

  /* configure */
  if (grantpt(fd) || unlockpt(fd)) goto err;

  /* user handle */
  mrsim->pty->fd = fd;
  if (device != mrsim->pty->device)
    snprintf(mrsim->pty->device, sizeof(mrsim->pty->device), "%s", device);
  mrsim->pty->r = mrsim->pty->w = 0;
  mrsim->pty->start = mrsim->pty->escape = false;

  return 0;
err:
  s = errno;
  close(fd);
  return s;
}


/* --- mrsim_close_pty ----------------------------------------------------- */

/* Close a pseudo terminal, remove symlink */
void
mrsim_close_pty(struct mrsim_pty_s *pty)
{
  if (pty->fd < 0) return;

  unlink(pty->device);
  close(pty->fd);
  pty->fd = -1;
}


/* --- mrsim_wait_msg ------------------------------------------------------ */

int
mrsim_wait_msg(const struct mrsim_s *mrsim, int timeoutms)
{
  struct pollfd pfds;
  int s;

  pfds.fd = mrsim->pty->fd;
  pfds.events = POLLIN;

  do {
    s = poll(&pfds, 1, timeoutms);
  } while (s < 0 && errno == EINTR);

  if (s <= 0) return s;

  if (pfds.revents & POLLIN) return 1;
  if (pfds.revents & POLLNVAL) return -1;

  if (pfds.revents & POLLHUP) {
    /* reopen */
    mrsim_close_pty(mrsim->pty);
    if (mrsim_open_pty(mrsim, mrsim->pty->device)) return -1;
    return 0;
  }

  return s;
}


/* --- mrsim_recv_msg ------------------------------------------------------ */

/* returns: 0: timeout/incomplete, -1: error, otherwise len of complete msg */

ssize_t
mrsim_recv_msg(struct mrsim_pty_s *pty, uint8_t **msg)
{
  struct iovec iov[2];
  ssize_t s;
  uint8_t c;

  *msg = NULL;
  do {
    /* feed the ring  buffer */
    iov[0].iov_base = pty->buf + pty->w;
    iov[1].iov_base = pty->buf;

    if (pty->r > pty->w) {
      iov[0].iov_len = pty->r - pty->w - 1;
      iov[1].iov_len = 0;
    } else if (pty->r > 0) {
      iov[0].iov_len = sizeof(pty->buf) - pty->w;
      iov[1].iov_len = pty->r - 1;
    } else {
      iov[0].iov_len = sizeof(pty->buf) - pty->w - 1;
      iov[1].iov_len = 0;
    }

    if (iov[0].iov_len || iov[1].iov_len) {
      do {
        s = readv(pty->fd, iov, 2);
      } while(s < 0 && errno == EINTR);

      if (s < 0) {
        if (pty->r == pty->w) return -1;
      } else if (s == 0) {
        if (pty->r == pty->w) return 0;
      } else
        pty->w = (pty->w + s) % sizeof(pty->buf);
    }

    while(pty->r != pty->w) {
      c = pty->buf[pty->r];
      pty->r = (pty->r + 1) % sizeof(pty->buf);

      switch(c) {
        case '^':
          pty->start = true;
          pty->escape = false;
          pty->len = 0;
          break;

        case '$':
          if (!pty->start) break;
          pty->start = false;
          *msg = pty->msg;
          return pty->len;

        case '!':
          pty->start = false;
          break;

        case '\\':
          pty->escape = true;
          break;

        default:
          if (!pty->start) break;
          if (pty->len >= sizeof(pty->msg)) {
            pty->start = false; break;
          }

          if (pty->escape) {
            c = ~c; pty->escape = false;
          }
          pty->msg[pty->len++] = c;
          break;
      }
    }
  } while(1);

  return 0;
}


/* --- mrsim_send_msg ------------------------------------------------------ */

static void	mrsim_encode(char x, char **buf);

int
mrsim_send_msg(const struct mrsim_pty_s *pty, const char *fmt, ...)
{
  va_list ap;
  ssize_t s;
  char buf[64], *w, *r;
  char c;

  if (pty->fd < 0) return -1;

  w = buf;

  va_start(ap, fmt);
  *w++ = '^';
  while((c = *fmt++)) {
    while (w - buf > (ssize_t)sizeof(buf)-8) {
      /* -8 = worst case (4 bytes escaped) */
      do {
        s = write(pty->fd, buf, w - buf);
      } while (s < 0 && errno == EINTR);
      if (s < 0) return -1;

      if (s > 0 && s < w - buf) memmove(buf, buf + s, w - buf - s);
      w -= s;
    }

    switch(c) {
      case '%': {
        switch(*fmt++) {
          case '1':
            mrsim_encode(va_arg(ap, int/*promotion*/), &w);
            break;

          case '2': {
            uint16_t x = va_arg(ap, int/*promotion*/);
            mrsim_encode((x >> 8) & 0xff, &w);
            mrsim_encode(x & 0xff, &w);
            break;
          }
          case '@': {
            uint16_t *x = va_arg(ap, uint16_t *);
            size_t l = va_arg(ap, size_t);
            while (l--) {
              mrsim_encode((*x >> 8) & 0xff, &w);
              mrsim_encode(*x & 0xff, &w);
              x++;
            }
            break;
          }

          case '4': {
            uint32_t x = va_arg(ap, uint32_t);
            mrsim_encode((x >> 24) & 0xff, &w);
            mrsim_encode((x >> 16) & 0xff, &w);
            mrsim_encode((x >> 8) & 0xff, &w);
            mrsim_encode(x & 0xff, &w);
            break;
          }
        }
        break;
      }

      default:
        mrsim_encode(c, &w);
    }
  }
  *w++ = '$';
  va_end(ap);

  r = buf;
  while (w > r) {
    do {
      s = write(pty->fd, r, w - r);
    } while (s < 0 && errno == EINTR);
    if (s < 0) return -1;

    r += s;
  }

  return 0;
}

static void
mrsim_encode(char x, char **buf)
{
  switch (x) {
    case '^': case '$': case '\\': case '!':
      *(*buf)++ = '\\';
      x = ~x;
  }
  *(*buf)++ = x;
}
